#Conversion of GenomeStudio output, pedigreefile and mapfile to Pedimap data file
#Roeland Voorrips, November 2011



GenomeStudioConvert <- function(mapfile, pedfile, GenomeStudioFile, PedimapFile, toNumbers=F) {
  map <- read.mapfile(mapfile)
  ped <- read.pedigreefile(pedfile)
  gsdata <- read.GenomeStudioFile(GenomeStudioFile)
  data <- convert.scores(gsdata,toNumbers)
  writePedimapFile(map,ped,data,PedimapFile)
}

#read.mapfile:
#mapfile is file name of tab-separated text file
#with 3 columns: marker, chromosome, position.
#there should be a caption line with at least these three captions.
#Any further columns are ignored and not retained
#Output is a corresponding data frame
#If an error occurs an attempt is made to interpret the file
#as a FlexQTL-style map file is read, see read.FQmapfile
read.mapfile <- function(mapfile) {
  map1 <- ""; map2 <- ""
  #first try to read as table with 3 columns:
  map1 <- tryCatch(
    { tmp <- read.table(mapfile, header=T, sep="\t", comment.char=";")
      if (class(tmp)!="data.frame") stop("No data.frame")
      captions <- tolower(names(tmp))
      if ( length(captions)<3 ||
           tolower(captions[1])!="marker" ||
           tolower(captions[2])!="chromosome" ||
           tolower(captions[3])!="position") {
        stop("Invalid captions")
      }
      names(tmp) <- captions #lower case
      #TODO: further tests for increasing position order ...
      subset(tmp, select=c(marker,chromosome,position)) #return of tryCatch if ok
    } , error=function(e) e$message)
  if (class(map1)!="character") {
    map1 #return value if table format
  }
  else {
    #not a table; try reading in FlexQTL / MapQTL format:
    map2 <- read.FQmapfile(mapfile)
    if (class(map2)!="character") {
      map2 #return value if FQ mapfile format
    }
    else {
      write(paste("Error reading",mapfile,":"),"")
      write(paste("when reading as table format: error message =",map1),"")
      write(paste("when reading as FlexQTL mapfile format: error message =",map2),"")
      stop("Aborted")
    }
  }  
} #read. mapfile

#read.FQmapfile:
#mapfile is the name of a text file containing a linkage map.
#each linkage group starts with a header line with the keyword "GROUP" or "CHROM"
#(upper/lower case not relevant) and the name of the group
#following the header lines are one or more lines, each with a locus name and position.
#The loci must be listed in ascending order of position (although that is not checked here)
#words and numbers must be separate by one or more spaces and/or tab characters
#Empty lines and lines starting with ";" are ignored 
#return value: error message (class character) is any error, 
#  else data.frame with columns marker, chromosome, position
read.FQmapfile <- function(mapfile) {
  tryCatch ( {
    line <- line <- readLines(mapfile, warn=F)
    i <- 1 #line number
    chromname <- ""
    loc <- 0
    locnames <- character(0)
    chrom <- character(0)
    position <- double(0)
    
    while (i<=length(line)) {
      words <- unlist(strsplit(line[i], split="[[:blank:]]"))
      words <- words[nchar(words)>0]
      words
      if (length(words)>0 && words[1][1] != ";") {
        W <- toupper(words[1])
        if (W=="GROUP" || W=="CHROM") {
          if (length(words)<2) stop (paste("GROUP or CHROM without group name on line",i))
          chromname <- words[2]
        }
        else {
          if (chromname=="") stop(paste(mapfile,": locus",words[1],"appears before GROUP header on line",i))
          if (length(words)<2 ||
             is.na(suppressWarnings(as.double(words[2]))) ) 
               stop (paste("locus name",words[1],"without valid position on line",i))
          #TODO: check if first locus or position >= previous locus on chrom 
          loc <- loc+1
          locnames[loc] <- words[1]
          chrom[loc] <- chromname
          position[loc] <- as.double(words[2])
        }
      } # line not empty or comment
      i <- i+1
    }  
    data.frame(marker=locnames, chromosome=chrom, position=position)
  }, error=function(e) e$message)  
} #read.FQmapfile

#read.pedigreefile:
#mapfile is file name of tab-separated text file
#with at least 3 columns: Individual, Parent1, Parent2
#any further columns are retained
#there should be a caption line, but its contents are ignored
#Output is a corresponding data frame
#note: pedigree not checked, errors only found in Pedimap
read.pedigreefile <- function(pedigreefile) {
    result <- read.table(pedigreefile, header=T, comment.char=";", na.strings="")
    if ( length(result)<3 ) {
      stop(paste("Less than 3 columns in ",pedigreefile))
    }
    result
} #read.pedigreefile

#read.GenomeStudioFile:
#GenomeStudioFile is file name of  tab-separated file
#We use only column Name for the marker names, and all columns <indiv>.GType
#for the individual genotypes
#The return value is a data frame with only the columns mentioned above,
#where the suffix ".GType has been removed from the individual names
read.GenomeStudioFile <- function(GenomeStudioFile) {
  all <- read.table(GenomeStudioFile, header=T, sep="\t", comment.char="")
  #Problem: individual names contain "-" and possibly other illegal characters
  #that are converted to dots.
  #possible solution: re-read the first line as a character vector
  #and export that seperately from the data frame
  con <- file(GenomeStudioFile, "r", blocking = FALSE)
  indnames <- unlist(strsplit(readLines(con, n=1),"\t")) 
  close(con)
  if (length(indnames) < length(all)) {
    indnames[length(all)] <- ""
  }
  endings <- substring(indnames,nchar(indnames)-5,nchar(indnames))
  gtypes <- toupper(endings)==".GTYPE" #vector of booleans
  selectcol <- (toupper(indnames)=="NAME") | gtypes
  gsdata <- subset(all, select=selectcol);
  indnames <- indnames[which(gtypes)]
  indnames <- substring(indnames,1,nchar(indnames)-6)
  names(gsdata) <- c("Name",indnames) #may result in illegal names, but we won't address gsdata by column names
  gsdata
} #read.GenomeStudioFile
 
getAlleleNames <- function(toNumbers,B.present) {
  if (B.present) {
    if (toNumbers) 41:42
    else c("A","B")
  }
  else {
    if (toNumbers) 51:54
    else c("A","C","G","T")
  }
}

#convert.scores takes a data frame as produced by read.GenomeStudioFile
#The genotypes can be either AA/AB/BB/--/NC (last two NoCalls) or
#any ATGC/--/NC letter combination
#If toNumbers==T we convert the letter code to numbers, for compatibility
#with FlexQTL
#AA/AB/BB -> A=41, B=42
#ACGT -> A=51, C=52, G=53, T=54
#The result is a list, with
# $toNumbers a boolean oh length 1
# $B.present a boolean of length 1, indicating if the scoring
#    was AA/AB/BB or in A/C/G/T letter combinations
# $scores a dataframe with a column $Name with the marker names
#  and two columns for each individual: $<indiv> and $indiv>_2
#  each column contains one of the two alleles (A/B/missing or A/C/T/G/missing,
#  or 41/42/missing or 51/52/53/54/missing)
convert.scores <- function(gsdata,toNumbers,missing="-") {
  nind <- length(gsdata)-1
  indnames <- names(gsdata)[2:length(gsdata)]
  new <- data.frame(Name=gsdata$Name)
  for (ind in 1:nind) {
    sc <- toupper(as.character(gsdata[[ind+1]]))
    #use -- for the no-call symbol:
    if ("NC" %in% sc) {
      sc[sc=="NC"] <- "--"
    }  
    # check if all codes have 2 characters:
    if (length(sc[nchar(sc)!=2])>0) {
      stop(paste("invalid code(s) at sample ",indnames[ind]))
    }
    # split codes into separate characters:
    new[[2*ind]] <- substring(sc,1,1)
    names(new)[2*ind] <- indnames[ind]
    new[2*ind+1] <- substring(sc,2,2)
    names(new)[2*ind+1] <- paste(indnames[ind],"_2",sep="")
  }
  # check scores
  B.present <- F
  i<-2                       
  while (!B.present && i<=length(new)) {
    B.present <- length(new[[i]][new[[i]]=="B"])>0                     
  }
  ok <- T
  i <- 2
  alnames <- c("-",getAlleleNames(toNumbers=F,B.present=B.present))
  while (ok && i<=length(new)) {
    ok <- length(new[[i]][!new[[i]] %in% alnames])==0
    i = i+1
  }
  if (!ok) stop(paste("invalid code(s) at sample ",names(new)[i]))
  if (toNumbers) {
    if (B.present) {
      for (i in 2:length(new)) {
        new[[i]][new[[i]]=="A"] <- 41
        new[[i]][new[[i]]=="B"] <- 42
        new[[i]][new[[i]]=="-"] <- missing
      }  
    }                       
    else {
      for (i in 2:length(new)) {
        new[[i]][new[[i]]=="A"] <- 51
        new[[i]][new[[i]]=="C"] <- 52
        new[[i]][new[[i]]=="G"] <- 53
        new[[i]][new[[i]]=="T"] <- 54
        new[[i]][new[[i]]=="-"] <- missing
      }  
    }
  }  
  else { # not toNumbers
    for (i in 2:length(new)) {
      new[[i]][new[[i]]=="-"] <- missing
    }
  }
  list(toNumbers=toNumbers,B.present=B.present,missing=missing,scores=new)
} #convert.scores 

# writePedimapFile takes the parameters:
# map: a dataframe with a linkage map as produced by read.mapfile
# ped: a dataframe with pedigree and possibly phenotypes as produced by read.pedigreefile
# data: a list as produced by score.convert
# filename: a new file (overwriting exiting ones) in Pedimap format.  
writePedimapFile <- function(map, ped, data, filename) {
  warnCount=0
  log = "GenomeStudioConvert.log"
  suppressWarnings( #suppress warning about deleting non-existing file
    file.remove(log)
  )
  #first the checks between the pedigree and GS data:
  GSindnames <- names(data$scores)
  GSindnames <- GSindnames[c(F,T)] #only keep first occurrence of each pair
  GSnotPed <- GSindnames[!GSindnames %in% ped[[1]]]
  if (length(GSnotPed)!=0) {
    if (length(GSnotPed)==length(GSindnames)) {
      stop("Error: none of the individuals in the GenomeStudio file are in the pedigree")
    }
    warnCount <- warnCount + 1
    write(paste(length(GSnotPed),"individuals in GenomeStudio file but not in pedigree"),"")
    write(paste(length(GSnotPed),"individuals in GenomeStudio file but not in pedigree:"),log, append=T)
    write.table(data.frame(x=GSnotPed),file=log,sep="\t",quote=F,col.names=F, row.names=F,append=T)
    #remove the unlisted individuals from data$scores:
    removeInd <- (1:length(GSindnames))[!(GSindnames %in% ped[[1]])]
    removeCols <- 2*removeInd #first column of each unmatched individual
    removeCols <- c(removeCols,(removeCols+1)) #include second columns
    data$scores <- subset(data$scores, select = -removeCols)
  }
  pedNotGS <- subset(ped, !(ped[[1]] %in% GSindnames), select=1)
  if (nrow(pedNotGS)!=0) {
    warnCount <- warnCount + 1
    write(paste(nrow(pedNotGS),"individuals in pedigree but not in GenomeStudio file"),"")
    write(paste(nrow(pedNotGS),"individuals in pedigree but not in GenomeStudio file:"),log, append=T)
    write.table(pedNotGS,file=log,sep="\t",quote=F,col.names=F, row.names=F,append=T)
  }
  #next the checks between the map and GS data
  GSmrknames <- as.character(data$scores$Name)
  map$marker <- as.character(map$marker)
  GSnotMap <- GSmrknames[!(GSmrknames %in% map$marker)]
  if (length(GSnotMap)!=0) {
    if (length(GSnotMap)==length(GSmrknames)) {
      suppressWarnings( #suppress warning about deleting non-existing file
        file.remove(log)
      )
      stop("Error: none of the markers in the GenomeStudio file are on the map")
    }
    warnCount <- warnCount + 1
    write(paste(length(GSnotMap),"markers in GenomeStudio file but not on map"),"")
    write(paste(length(GSnotMap),"markers in GenomeStudio file but not on map:"),log, append=T)
    write.table(data.frame(x=GSnotMap),file=log,sep="\t",quote=F,col.names=F, row.names=F,append=T)
    #not necessary to remove the unlisted markers from data$scores, they won't be written
  }
  mapNotGS <- subset(map, !(map$marker %in% GSmrknames))
  if (nrow(mapNotGS)!=0) {
    warnCount <- warnCount + 1
    write(paste(nrow(mapNotGS),"markers in map but not in GenomeStudio file"),"")
    write(paste(nrow(mapNotGS),"markers in map but not in GenomeStudio file:"),log, append=T)
    write.table(mapNotGS,file=log,sep="\t",quote=F,col.names=F, row.names=F,append=T)
    #remove unlisted markers from map:
    map <- subset(map, marker %in% GSmrknames)
  }
  if (warnCount>0) {
    write(paste("See file",log,"for non-matching individuals and/or markers"),"")
  }

  # write header lines:
  write("POPULATION = GenomeStudio_Import", file=filename)
  write(paste("UNKNOWN = ",data$missing),file=filename, append=T)
  write("PLOIDY = 2",file=filename,append=T)
  
  # write pedigree and possible phenotype columns
  write("",file=filename,append=T)
  write("PEDIGREE",file=filename,append=T)
  write("",file=filename,append=T)
  suppressWarnings( #suppress warning about appending column names
    write.table(ped,file=filename,sep="\t",quote=F,col.names=T, row.names=F,append=T)
  )
  
  #get the list of chromosome names in supplied order (not alphabetical / numerical order)
  map$chromosome <- as.character(map$chromosome)
  chromnames <- c(map$chromosome,"")
  chromlist2 <- c("",map$chromosome)
  chromnames <- chromnames[chromnames!=chromlist2]
  length(chromnames) <- length(chromnames)-1 
  #write linkage groups
  for (c in 1:length(chromnames)) {
    write("",file=filename,append=T)
    write(paste("LINKAGEGROUP ",chromnames[c]),file=filename,append=T)
    write("",file=filename,append=T)
    write("MAP",file=filename,append=T)
    chrmap <- subset(map, chromosome==chromnames[c], select = c(marker,position))
    write.table(chrmap,file=filename,sep="\t",quote=F,col.names=F,row.names=F,append=T)
    for (loc in 1:length(chrmap$marker)) {
      write("",file=filename,append=T)
      write(paste("LOCUS ",chrmap$marker[loc]),file=filename,append=T)
      write(c("ALLELENAMES",getAlleleNames(data$toNumbers,data$B.present)),file=filename,ncolumns=4,sep="\t",append=T)
    }
  } # for c
  
  # write allele scores
  GSindnames <- names(data$scores) #again, as individuals not in the pedigree may have been removed
  GSindnames <- GSindnames[c(F,T)] #only keep first occurrence of each pair
  for (loc in 1:length(map$marker)) {
    sc <- subset(data$scores[],Name==map$marker[loc])
    if (nrow(sc)>0) { #else marker not in map, don't write
      write("",file=filename,append=T)
      write(paste("ALLELES",map$marker[loc],"; pos",map$chromosome[loc],map$position[loc], sep=" "),
            file=filename,append=T)
      sc <- sc[2:length(sc)] #without marker name: 2*nind scores
      scmatrix <- matrix(sc,ncol=2,nrow=length(sc)/2,byrow=T)
      rownames(scmatrix) <- GSindnames
      write.table(scmatrix,file=filename,sep="\t",quote=F,col.names=F,row.names=T,append=T)
    }  
  } # for loc
} #writePedimapFile

    
################ here the actual calls #################################################
args <- commandArgs(trailingOnly = TRUE)
write("", "")
write("Parameters :", "")
write(paste("map file =           ",args[1]), "")
write(paste("pedigree file =      ",args[2]), "")
write(paste("GenomeStudio file =  ",args[3]), "")
write(paste("Pedimap output file =",args[4]), "")
#print(args)
if (length(args)<4) {
  write("", "")
  write("at least 4 parameters are required: mapfile, pedigreefile, GenomeStudioFile, PedimapFile", "")
} else {
  toNumbers <- F
  if ( (length(args)>4) &&
       (toupper(args[5]) %in% c("1","YES","TRUE")) )  {
    toNumbers <- T
    write("convert allele identifiers to numbers", "")
  }
  write("", "")
  GenomeStudioConvert(args[1],args[2],args[3],args[4],toNumbers)  
  write("ready", "")
}  
